package ru.dbotthepony.mc.otm.capability

import java.util.function.Predicate

/**
 * Represents possible flow direction, both for matter and for energy
 *
 * To dynamically get this enum by two booleans (by input and output states), use [FlowDirection.of]
 *
 * Also, this enum implements [Predicate], you can check incoming flow direction for "compatibility" with
 * this flow direction, more precisely:
 * [test] will return true and only true if [input] and [output] of incoming flow are assignable to this flow
 *
 * Example:
 * * `INPUT.test(OUTPUT)` = `false`
 * * `OUTPUT.test(INPUT)` = `false`
 * * `INPUT.test(BI_DIRECTIONAL)` = `true`
 * * `OUTPUT.test(BI_DIRECTIONAL)` = `true`
 * * `BI_DIRECTIONAL.test(BI_DIRECTIONAL)` = `true`
 * * `BI_DIRECTIONAL.test(OUTPUT)` = `false`
 * * `BI_DIRECTIONAL.test(INPUT)` = `false`
 */
enum class FlowDirection(val input: Boolean, val output: Boolean) : Predicate<FlowDirection> {
	/**
	 * Can only be inputted (consumer)
	 */
	INPUT(true, false),

	/**
	 * Can only be outputted/transmitted (producer)
	 */
	OUTPUT(false, true),

	/**
	 * Can both consume and produce (capacitor)
	 */
	BI_DIRECTIONAL(true, true),

	/**
	 * Why would you want to use this
	 */
	NONE(false, false);

	override fun test(t: FlowDirection): Boolean {
		return t === this || (!input || t.input) && (!output || t.output)
	}

	fun intersect(other: FlowDirection): FlowDirection {
		return of(other.input && input, other.output && output)
	}

	companion object {
		@JvmStatic
		fun of(input: Boolean, output: Boolean): FlowDirection {
			if (input && output) {
				return BI_DIRECTIONAL
			} else if (input) {
				return INPUT
			} else if (output) {
				return OUTPUT
			} else {
				return NONE
			}
		}

		/**
		 * @return [INPUT] if [flag] is true, [OUTPUT] otherwise
		 */
		@JvmStatic
		fun input(flag: Boolean): FlowDirection {
			return of(flag, !flag)
		}

		/**
		 * @return [OUTPUT] if [flag] is true, [INPUT] otherwise
		 */
		@JvmStatic
		fun output(flag: Boolean): FlowDirection {
			return of(!flag, flag)
		}

		/**
		 * @return [INPUT] if [flag] is true, [NONE] otherwise
		 */
		@JvmStatic
		fun orInput(flag: Boolean): FlowDirection {
			return of(flag, false)
		}

		/**
		 * @return [OUTPUT] if [flag] is true, [NONE] otherwise
		 */
		@JvmStatic
		fun orOutput(flag: Boolean): FlowDirection {
			return of(false, flag)
		}
	}
}
